"""
Tests the higher level functions in bet_sizing.py.
"""

import unittest
from unittest.mock import patch
import datetime as dt
import numpy as np
import pandas as pd
from scipy.stats import norm, moment

from famlafl.bet_sizing.bet_sizing import (bet_size_probability, bet_size_dynamic, bet_size_budget, bet_size_reserve,
                                            confirm_and_cast_to_df, get_concurrent_sides, cdf_mixture,
                                            single_bet_size_mixed)
from famlafl.bet_sizing.ch10_snippets import (get_signal, avg_active_signals, discrete_signal,
                                               get_w, get_target_pos, limit_price, bet_size)
from famlafl.bet_sizing.ef3m import M2N, raw_moment, most_likely_parameters


class TestBetSizeProbability(unittest.TestCase):
    """
    Tests the 'bet_size_probability' function.
    """
    def test_bet_size_probability_default(self):
        """
        Tests for successful execution using the default arguments of 'bet_size_probability', which are:
         average_active = False
         step_size = 0.0
        """
        # Setup the test DataFrame.
        dates_test = np.array([dt.datetime(2000, 1, 1) + i * dt.timedelta(days=1) for i in range(5)])
        shift_dt = np.array([dt.timedelta(days=0.5*i+1) for i in range(5)])
        dates_shifted_test = dates_test + shift_dt
        events_test = pd.DataFrame(data=[[0.55, 1], [0.7, -1], [0.95, 1], [0.65, -1], [0.85, 1]],
                                   columns=['prob', 'side'],
                                   index=dates_test)
        events_test['t1'] = dates_shifted_test
        # Calculate correct output.
        signal_0 = get_signal(events_test['prob'], 2, events_test['side'])
        df_signal_0 = signal_0.to_frame('signal').join(events_test['t1'], how='left')
        signal_1 = df_signal_0.signal
        # Evaluate test.
        self.assertTrue(signal_1.equals(bet_size_probability(events_test, events_test['prob'], 2, events_test['side'])))

    def test_bet_size_probability_avg_active(self):
        """
        Tests for successful execution of 'bet_size_probability' with 'average_active' set to True.
        """
        # Setup the test DataFrame.
        dates_test = np.array([dt.datetime(2000, 1, 1) + i * dt.timedelta(days=1) for i in range(5)])
        shift_dt = np.array([dt.timedelta(days=0.5*i+1) for i in range(5)])
        dates_shifted_test = dates_test + shift_dt
        events_test = pd.DataFrame(data=[[0.55, 1], [0.7, -1], [0.95, 1], [0.65, -1], [0.85, 1]],
                                   columns=['prob', 'side'],
                                   index=dates_test)
        events_test['t1'] = dates_shifted_test
        # Calculate correct output.
        signal_0 = get_signal(events_test['prob'], 2, events_test['side'])
        df_signal_0 = signal_0.to_frame('signal').join(events_test['t1'], how='left')
        signal_1 = avg_active_signals(df_signal_0, 1)
        # Evaluate test.
        self.assertTrue(signal_1.equals(bet_size_probability(events=events_test, prob=events_test['prob'], num_classes=2,
                                                             pred=events_test['side'], average_active=True)))

    def test_bet_size_probability_stepsize(self):
        """
        Tests for successful execution of 'bet_size_probability' with 'step_size' greater than 0.
        """
        # Setup the test DataFrame.
        dates_test = np.array([dt.datetime(2000, 1, 1) + i * dt.timedelta(days=1) for i in range(5)])
        shift_dt = np.array([dt.timedelta(days=0.5*i+1) for i in range(5)])
        dates_shifted_test = dates_test + shift_dt
        events_test = pd.DataFrame(data=[[0.55, 1], [0.7, -1], [0.95, 1], [0.65, -1], [0.85, 1]],
                                   columns=['prob', 'side'],
                                   index=dates_test)
        events_test['t1'] = dates_shifted_test
        # Calculate correct output.
        signal_0 = get_signal(events_test['prob'], 2, events_test['side'])
        df_signal_0 = signal_0.to_frame('signal').join(events_test['t1'], how='left')
        signal_1 = df_signal_0.signal
        signal_1 = discrete_signal(signal0=signal_1, step_size=0.1)
        # Evaluate test.
        self.assertTrue(signal_1.equals(bet_size_probability(events=events_test, prob=events_test['prob'], num_classes=2,
                                                             pred=events_test['side'], step_size=0.1)))


class TestBetSizeDynamic(unittest.TestCase):
    """
    Tests the 'bet_size_dynamic' function.
    """
    def test_bet_size_dynamic_default(self):
        """
        Tests for successful execution using the default arguments of 'bet_size_dynamic', which are:
         average_active = False
         step_size = 0.0
        """
        # Setup the test DataFrame.
        dates_test = np.array([dt.datetime(2000, 1, 1) + i * dt.timedelta(days=1) for i in range(5)])
        events_test = pd.DataFrame(data=[[25, 55, 75.50, 80.00],
                                         [35, 55, 76.90, 75.00],
                                         [45, 55, 74.10, 72.50],
                                         [40, 55, 67.75, 65.00],
                                         [30, 55, 62.00, 70.80]],
                                   columns=['pos', 'max_pos', 'm_p', 'f'],
                                   index=dates_test)
        # Calculate results.
        d_events = {col: events_test[col] for col in list(events_test.columns)}
        events_results = confirm_and_cast_to_df(d_events)
        w_param = get_w(10, 0.95, 'sigmoid')
        events_results['t_pos'] = events_results.apply(lambda row: get_target_pos(w_param, row.f, row.m_p,
                                                                                  row.max_pos, 'sigmoid'), axis=1)
        events_results['l_p'] = events_results.apply(lambda row: limit_price(row.t_pos, row.pos, row.f, w_param,
                                                                             row.max_pos, 'sigmoid'), axis=1)
        events_results['bet_size'] = events_results.apply(lambda row: bet_size(w_param, row.f-row.m_p, 'sigmoid'),
                                                          axis=1)
        df_result = events_results[['bet_size', 't_pos', 'l_p']]
        # Evaluate.
        self.assertTrue(df_result.equals(bet_size_dynamic(events_test['pos'], events_test['max_pos'],
                                                          events_test['m_p'], events_test['f'])))


class TestBetSizeBudget(unittest.TestCase):
    """
    Tests the 'bet_size_budget' function.
    """
    def test_bet_size_budget_default(self):
        """
        Tests for the successful execution of the 'bet_size_budget' function.
        """
        # Setup the test DataFrame.
        dates_test = np.array([dt.datetime(2000, 1, 1) + i * dt.timedelta(days=1) for i in range(5)])
        shift_dt = np.array([dt.timedelta(days=0.5*i+1) for i in range(5)])
        dates_shifted_test = dates_test + shift_dt
        events_test = pd.DataFrame(data=[[0.55, 1], [0.7, 1], [0.95, 1], [0.65, -1], [0.85, 1]],
                                   columns=['prob', 'side'],
                                   index=dates_test)
        events_test['t1'] = dates_shifted_test
        # Calculate correct result.
        events_result = get_concurrent_sides(events_test['t1'], events_test['side'])
        avg_long = events_result['active_long'] / events_result['active_long'].max()
        avg_short = events_result['active_short'] / events_result['active_short'].max()
        events_result['bet_size'] = avg_long - avg_short
        # Evaluate.
        self.assertTrue(events_result.equals(bet_size_budget(events_test['t1'], events_test['side'])))

    def test_bet_size_budget_div_zero(self):
        """
        Tests for successful handling of events DataFrames that result in a maximum number of
        concurrent bet sides of zero.
        """
        # Setup the test DataFrame.
        dates_test = np.array([dt.datetime(2000, 1, 1) + i * dt.timedelta(days=1) for i in range(5)])
        shift_dt = np.array([dt.timedelta(days=0.5*i+1) for i in range(5)])
        dates_shifted_test = dates_test + shift_dt
        events_test = pd.DataFrame(data=[[0.55, 1], [0.7, 1], [0.95, 1], [0.65, 1], [0.85, 1]],
                                   columns=['prob', 'side'],
                                   index=dates_test)
        events_test['t1'] = dates_shifted_test
        # Calculate correct results.
        events_result = get_concurrent_sides(events_test['t1'], events_test['side'])
        max_active_long, max_active_short = events_result['active_long'].max(), events_result['active_short'].max()
        avg_long = events_result['active_long'] / max_active_long if max_active_long > 0 else 0
        avg_short = events_result['active_short'] / max_active_short if max_active_short > 0 else 0
        events_result['bet_size'] = avg_long - avg_short
        # Evaluate.
        self.assertTrue(events_result.equals(bet_size_budget(events_test['t1'], events_test['side'])))


class TestBetSizeReserve(unittest.TestCase):
    """
    Tests the 'bet_size_reserve' function.
    """
    @patch('famlafl.bet_sizing.bet_sizing.most_likely_parameters')
    def test_bet_size_reserve_default(self, mock_likely_parameters):
        """
        Tests for successful execution of 'bet_size_reserve' using default arguments, return_parameters=False.
        Function 'most_likely_parameters' needs to be patched because the 'M2N.mp_fit' method makes use of
        random numbers.
        """
        # Setup the test DataFrame.
        np.random.seed(0)
        sample_size = 500
        start_date = dt.datetime(2000, 1, 1)
        date_step = dt.timedelta(days=1)
        dates = np.array([start_date + i*date_step for i in range(sample_size)])
        shift_dt = np.array([dt.timedelta(days=d) for d in np.random.uniform(1., 20., sample_size)])
        dates_shifted = dates + shift_dt
        time_1 = pd.Series(data=dates_shifted, index=dates)
        df_events = time_1.to_frame()
        df_events = df_events.rename(columns={0: 't1'})
        df_events['p'] = np.random.uniform(0.0, 1.0, sample_size)
        df_events = df_events[['t1', 'p']]
        df_events['side'] = df_events['p'].apply(lambda x: 1 if x >= 0.5 else -1)
        # Calculate the correct results.
        events_active = get_concurrent_sides(df_events['t1'], df_events['side'])
        events_active['c_t'] = events_active['active_long'] - events_active['active_short']
        central_moments = [moment(events_active['c_t'].to_numpy(), moment=i) for i in range(1, 6)]
        raw_moments = raw_moment(central_moments=central_moments, dist_mean=events_active['c_t'].mean())
        m2n_test = M2N(raw_moments, epsilon=1e-5, factor=5, n_runs=10, variant=2, max_iter=10_000, num_workers=1)
        test_results = m2n_test.mp_fit()
        test_params = most_likely_parameters(test_results)
        mock_likely_parameters.return_value = test_params
        test_fit = [test_params[key] for key in ['mu_1', 'mu_2', 'sigma_1', 'sigma_2', 'p_1']]
        events_active['bet_size'] = events_active['c_t'].apply(lambda c: single_bet_size_mixed(c, test_fit))
        # Evaluate.
        df_bet = bet_size_reserve(df_events['t1'], df_events['side'], fit_runs=10)
        self.assertTrue(events_active.equals(df_bet))

    @patch('famlafl.bet_sizing.bet_sizing.most_likely_parameters')
    def test_bet_size_reserve_return_params(self, mock_likely_parameters):
        """
        Tests for successful execution of 'bet_size_reserve' using return_parameters=True.
        Function 'most_likely_parameters' needs to be patched because the 'M2N.mp_fit' method makes use of
        random numbers.
        """
        # Setup the test DataFrame.
        np.random.seed(0)
        sample_size = 500
        start_date = dt.datetime(2000, 1, 1)
        date_step = dt.timedelta(days=1)
        dates = np.array([start_date + i*date_step for i in range(sample_size)])
        shift_dt = np.array([dt.timedelta(days=d) for d in np.random.uniform(1., 20., sample_size)])
        dates_shifted = dates + shift_dt
        time_1 = pd.Series(data=dates_shifted, index=dates)
        df_events = time_1.to_frame()
        df_events = df_events.rename(columns={0: 't1'})
        df_events['p'] = np.random.uniform(0.0, 1.0, sample_size)
        df_events = df_events[['t1', 'p']]
        df_events['side'] = df_events['p'].apply(lambda x: 1 if x >= 0.5 else -1)
        # Calculate the correct results.
        events_active = get_concurrent_sides(df_events['t1'], df_events['side'])
        events_active['c_t'] = events_active['active_long'] - events_active['active_short']
        central_moments = [moment(events_active['c_t'].to_numpy(), moment=i) for i in range(1, 6)]
        raw_moments = raw_moment(central_moments=central_moments, dist_mean=events_active['c_t'].mean())
        m2n_test = M2N(raw_moments, epsilon=1e-5, factor=5, n_runs=10, variant=2, max_iter=10_000, num_workers=1)
        test_results = m2n_test.mp_fit()
        test_params = most_likely_parameters(test_results)
        mock_likely_parameters.return_value = test_params
        test_fit = [test_params[key] for key in ['mu_1', 'mu_2', 'sigma_1', 'sigma_2', 'p_1']]
        events_active['bet_size'] = events_active['c_t'].apply(lambda c: single_bet_size_mixed(c, test_fit))
        # Evaluate.
        eval_events, eval_params = bet_size_reserve(df_events['t1'], df_events['side'],
                                                    fit_runs=10, return_parameters=True)
        self.assertEqual(test_params, eval_params)
        self.assertTrue(events_active.equals(eval_events))


class TestConfirmAndCastToDf(unittest.TestCase):
    """
    Tests the 'confirm_and_cast_to_df' function.
    """
    def test_cast_to_df_all_series(self):
        """
        Tests for successful execution of 'confirm_and_cast_to_df' when all dictionary values are pandas.Series.
        """
        # Setup the test DataFrame.
        dates_test = np.array([dt.datetime(2000, 1, 1) + i * dt.timedelta(days=1) for i in range(5)])
        events_test = pd.DataFrame(data=[[25, 55, 75.50, 80.00],
                                         [35, 55, 76.90, 75.00],
                                         [45, 55, 74.10, 72.50],
                                         [40, 55, 67.75, 65.00],
                                         [30, 55, 62.00, 70.80]],
                                   columns=['pos', 'max_pos', 'm_p', 'f'],
                                   index=dates_test)
        d_events = {col: events_test[col] for col in list(events_test.columns)}
        # Evaluate.
        df_results = confirm_and_cast_to_df(d_events)
        self.assertTrue(events_test.equals(df_results))

    def test_cast_to_df_one_series(self):
        """
        Tests for successful execution of 'confirm_and_cast_to_df' when only one of the dictionary values are
        a pandas.Series.
        """
        # Setup the test DataFrame.
        dates_test = np.array([dt.datetime(2000, 1, 1) + i * dt.timedelta(days=1) for i in range(5)])
        max_pos, market_price, forecast_price = 55, 75.00, 80.00
        events_test = pd.DataFrame(data=[[25, max_pos, market_price, forecast_price],
                                         [35, max_pos, market_price, forecast_price],
                                         [45, max_pos, market_price, forecast_price],
                                         [40, max_pos, market_price, forecast_price],
                                         [30, max_pos, market_price, forecast_price]],
                                   columns=['pos', 'max_pos', 'm_p', 'f'],
                                   index=dates_test)
        d_events = {'pos': events_test['pos'], 'max_pos': max_pos, 'm_p': market_price, 'f': forecast_price}
        # Evaluate.
        df_results = confirm_and_cast_to_df(d_events)
        self.assertTrue(np.allclose(events_test.to_numpy(), df_results.to_numpy(), 1e-9))

    def test_cast_to_df_no_series(self):
        """
        Tests for successful execution of 'confirm_and_cast_to_df' when none of the dictionary values are
        a pandas.Series.
        """
        # Setup the test DataFrame.
        pos, max_pos, market_price, forecast_price = 35, 55, 75.00, 80.00
        events_test = pd.DataFrame(data=[[pos, max_pos, market_price, forecast_price]],
                                   columns=['pos', 'max_pos', 'm_p', 'f'])
        d_events = {'pos': pos, 'max_pos': max_pos, 'm_p': market_price, 'f': forecast_price}
        # Evaluate.
        df_results = confirm_and_cast_to_df(d_events)
        self.assertTrue(np.allclose(events_test.to_numpy(), df_results.to_numpy(), 1e-9))


class TestGetConcurrentSides(unittest.TestCase):
    """
    Tests the function 'get_concurrent_sides' for successful operation.
    """
    def test_get_concurrent_sides_default(self):
        """
        Tests for the successful execution of 'get_concurrent_sides'. Since there are no options or branches,
        there are no additional test cases beyond default.
        """
        # Setup the test DataFrame.
        np.random.seed(0)
        sample_size = 100
        start_date = dt.datetime(2000, 1, 1)
        date_step = dt.timedelta(days=1)
        dates = np.array([start_date + i*date_step for i in range(sample_size)])
        shift_dt = np.array([dt.timedelta(days=d) for d in np.random.uniform(1., 20., sample_size)])
        dates_shifted = dates + shift_dt
        time_1 = pd.Series(data=dates_shifted, index=dates)
        df_events = time_1.to_frame()
        df_events = df_events.rename(columns={0: 't1'})
        df_events['p'] = np.random.uniform(0.0, 1.0, sample_size)
        df_events = df_events[['t1', 'p']]
        df_events['side'] = df_events['p'].apply(lambda x: 1 if x >= 0.5 else -1)
        # Calculate correct result.
        events_test = df_events.copy()
        events_test['active_long'] = 0
        events_test['active_short'] = 0
        for idx in events_test.index:
            df_long_active_idx = set(events_test[(events_test.index <= idx) & (events_test['t1'] > idx) \
                                        & (events_test['side'] > 0)].index)
            events_test.loc[idx, 'active_long'] = len(df_long_active_idx)
            df_short_active_idx = set(events_test[(events_test.index <= idx) & (events_test['t1'] > idx) \
                                        & (events_test['side'] < 0)].index)
            events_test.loc[idx, 'active_short'] = len(df_short_active_idx)
        events_test = events_test[['active_long', 'active_short']]
        # Evaluate.
        df_results = get_concurrent_sides(df_events['t1'], df_events['side'])
        self.assertTrue(np.allclose(events_test.to_numpy(), df_results[['active_long', 'active_short']].to_numpy(), 1e-9))


class TestCdfMixture(unittest.TestCase):
    """
    Tests the 'cdf_mixture' function.
    """
    def test_cdf_mixture_default(self):
        """
        Tests for the successful execution of the 'cdf_mixture' function. Since there are no options or branches,
        there are no additional test cases beyond default.
        """
        # Setup the test data.
        x_value = 5.0
        mu_1, mu_2, sigma_1, sigma_2, p_1 = -1.0, 4.0, 2.0, 1.5, 0.4
        params = [mu_1, mu_2, sigma_1, sigma_2, p_1]
        # Calculate the expected results.
        expected_result = (p_1 * norm.cdf(x_value, mu_1, sigma_1)) + ((1-p_1) * norm.cdf(x_value, mu_2, sigma_2))
        # Evaluate.
        self.assertEqual(expected_result, cdf_mixture(x_value, params))


class TestSingleBetSizeMixed(unittest.TestCase):
    """
    Tests the 'single_bet_size_mixed' function.
    """
    def test_single_bet_size_mixed_above_zero(self):
        """
        Tests for the successful execution of the 'single_bet_size_mixed' function where the 'c_t' parameter is
        greater than zero.
        """
        # Setup the test data.
        c_t = 5.0
        mu_1, mu_2, sigma_1, sigma_2, p_1 = -1.0, 4.0, 2.0, 1.5, 0.4
        params = [mu_1, mu_2, sigma_1, sigma_2, p_1]
        # Calculate the expected result.
        expected_result = (cdf_mixture(c_t, params) - cdf_mixture(0, params)) / (1 - cdf_mixture(0, params))
        # Evaluate.
        self.assertEqual(expected_result, single_bet_size_mixed(c_t, params))

    def test_single_bet_size_mixed_below_zero(self):
        """
        Tests for the successful execution of the 'single_bet_size_mixed' function where the 'c_t' parameter is
        less than zero.
        """
        # Setup the test data.
        c_t = -4.0
        mu_1, mu_2, sigma_1, sigma_2, p_1 = -1.0, 4.0, 2.0, 1.5, 0.4
        params = [mu_1, mu_2, sigma_1, sigma_2, p_1]
        # Calculate the expected result.
        expected_result = (cdf_mixture(c_t, params) - cdf_mixture(0, params)) / cdf_mixture(0, params)
        # Evaluate.
        self.assertEqual(expected_result, single_bet_size_mixed(c_t, params))



"""
Tests Follow the Leader.
"""
from unittest import TestCase
import os
import numpy as np
import pandas as pd
from famlafl.online_portfolio_selection.ftl import FTL


class TestFollowTheLeader(TestCase):
    # pylint: disable=too-many-public-methods
    # pylint: disable=unsubscriptable-object
    # pylint: disable=protected-access
    """
    Tests different functions of the Follow the Leader class.
    """

    def setUp(self):
        """
        Sets the file path for the tick data csv.
        """
        # Set project path to current directory.
        project_path = os.path.dirname(__file__)
        # Add new data path to match stock_prices.csv data.
        data_path = project_path + '/test_data/stock_prices.csv'
        # Read csv, parse dates, and drop NaN.
        self.data = pd.read_csv(data_path, parse_dates=True, index_col="Date").dropna(axis=1)

    def test_ftl_solution(self):
        """
        Test the calculation of follow the leader.
        """
        # Initialize FTL.
        ftl = FTL()
        # Allocate asset prices to FTL.
        ftl.allocate(self.data, resample_by='M')
        all_weights = np.array(ftl.all_weights)
        for i in range(all_weights.shape[0]):
            weights = all_weights[i]
            assert (weights >= 0).all()
            assert len(weights) == self.data.shape[1]
            np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_ftl_first_weight(self):
        """
        Tests that the weights calculated for the first time period is uniform.
        """
        # Initialize FTL.
        ftl1 = FTL()
        # Allocate asset prices to FTL.
        ftl1.allocate(self.data, resample_by='M')
        all_weights = np.array(ftl1.all_weights)
        # Get uniform weights.
        uniform_weight = ftl1._uniform_weight()
        # Compare the first weight to uniform weights.
        np.testing.assert_almost_equal(uniform_weight, all_weights[0])



"""
Contains the code for implementing sample weights.
"""

from famlafl.sample_weights.attribution import (get_weights_by_time_decay, get_weights_by_return,
                                                 _apply_weight_by_return)



"""
Test Optimal Number of Clusters algorithm
"""

import unittest
import pandas as pd
from sklearn.datasets import load_breast_cancer

from famlafl.clustering import get_onc_clusters
from famlafl.clustering.onc import _check_improve_clusters


class TestOptimalNumberOfClusters(unittest.TestCase):
    """
    Test get_onc_clusters function
    """

    def setUp(self):
        """
        Set the file path for the sample dollar bars data.
        """
        self.data, _ = load_breast_cancer(return_X_y=True)

    @staticmethod
    def _check_if_in_cluster(array, cluster_dict):
        """
        Check if array is in dictionary values not taking into account order of elements
        """
        array_set = set(array)
        flag = False
        for arr in cluster_dict.values():
            if set(arr) == array_set:
                flag = True
        return flag

    def test_get_onc_clusters(self):
        """
        Test get_onc_clusters function on Breast Cancer data set from sklearn
        """

        _, clusters, _ = get_onc_clusters(pd.DataFrame(self.data).corr(), repeat=50)
        self.assertGreaterEqual(len(clusters.keys()), 5)  # Optimal number of clusters
        self.assertTrue(self._check_if_in_cluster([11, 14, 18], clusters))  # Check clusters components
        self.assertTrue(self._check_if_in_cluster([0, 2, 3, 10, 12, 13, 20, 22, 23], clusters))
        self.assertTrue(self._check_if_in_cluster([5, 6, 7, 25, 26, 27], clusters))

    def test_check_redo_condition(self):
        """
        Function to test redo condition helper function
        """
        self.assertEqual((4, 5, 6), _check_improve_clusters(2, 3, (1, 2, 3), (4, 5, 6)))



"""
Test structural break tests: Chow-type, CUSUM, SADF
"""

import unittest
import os
import numpy as np
import pandas as pd

from famlafl.structural_breaks import (get_chow_type_stat, get_sadf, get_chu_stinchcombe_white_statistics)


# pylint: disable=unsubscriptable-object
from famlafl.structural_breaks.cusum import _get_values_diff
from famlafl.structural_breaks.sadf import get_betas


class TesStructuralBreaks(unittest.TestCase):
    """
    Test Chow-type, CUSUM, SADF tests
    """

    def setUp(self):
        """
        Set the file path for the sample dollar bars data.
        """
        project_path = os.path.dirname(__file__)
        self.path = project_path + '/test_data/dollar_bar_sample.csv'
        self.data = pd.read_csv(self.path, index_col='date_time', parse_dates=[0])

    def test_chow_test(self):
        """
        Test get_chow_type_stat function
        """
        min_length = 10
        log_prices = np.log(self.data.close)
        stats = get_chow_type_stat(log_prices, min_length=min_length, verbose=True)

        # We drop first and last # of min_length values
        self.assertEqual(log_prices.shape[0] - min_length * 2, stats.shape[0])
        self.assertAlmostEqual(stats.max(), 0.179, delta=1e-3)
        self.assertAlmostEqual(stats.mean(), -0.653, delta=1e-3)
        self.assertAlmostEqual(stats[3], -0.6649, delta=1e-3)

    def test_chu_stinchcombe_value_diff_function(self):
        """
        Test the values diff hidden function.
        """
        # Test values diff function
        one_sided_diff = _get_values_diff(test_type='one_sided', series=pd.Series([1, 2, 3, 4, 5]), index=0, ind=1)
        two_sided_diff = _get_values_diff(test_type='two_sided', series=pd.Series([1, 2, 3, 4, 5]), index=0, ind=1)
        self.assertEqual(-1, one_sided_diff)
        self.assertEqual(1, two_sided_diff)
        self.assertRaises(ValueError, _get_values_diff, test_type='rubbish',
                          series=pd.Series([1, 2, 3, 4, 5]), index=0, ind=1)

    def test_chu_stinchcombe_white_test(self):
        """
        Test get_chu_stinchcombe_white_statistics function
        """

        log_prices = np.log(self.data.close)
        one_sided_test = get_chu_stinchcombe_white_statistics(log_prices, test_type='one_sided', verbose=True)
        two_sided_test = get_chu_stinchcombe_white_statistics(log_prices, test_type='two_sided', verbose=True)

        # For the first two values we don't have enough info
        self.assertEqual(log_prices.shape[0] - 2, one_sided_test.shape[0])
        self.assertEqual(log_prices.shape[0] - 2, two_sided_test.shape[0])

        self.assertAlmostEqual(one_sided_test.critical_value.max(), 3.265, delta=1e-3)
        self.assertAlmostEqual(one_sided_test.critical_value.mean(), 2.7809, delta=1e-3)
        self.assertAlmostEqual(one_sided_test.critical_value[20], 2.4466, delta=1e-3)

        self.assertAlmostEqual(one_sided_test.stat.max(), 3729.001, delta=1e-3)
        self.assertAlmostEqual(one_sided_test.stat.mean(), 836.509, delta=1e-3)
        self.assertAlmostEqual(one_sided_test.stat[20], 380.137, delta=1e-3)

        self.assertAlmostEqual(two_sided_test.critical_value.max(), 3.235, delta=1e-3)
        self.assertAlmostEqual(two_sided_test.critical_value.mean(), 2.769, delta=1e-3)
        self.assertAlmostEqual(two_sided_test.critical_value[20], 2.715, delta=1e-3)

        self.assertAlmostEqual(two_sided_test.stat.max(), 5518.519, delta=1e-3)
        self.assertAlmostEqual(two_sided_test.stat.mean(), 1264.582, delta=1e-3)
        self.assertAlmostEqual(two_sided_test.stat[20], 921.2979, delta=1e-3)

        self.assertRaises(ValueError, get_chu_stinchcombe_white_statistics, log_prices, 'rubbish text')

    def test_sadf_test(self):
        """
        Test get_sadf function
        """

        log_prices = np.log(self.data.close)
        lags_int = 5
        lags_array = [1, 2, 5, 7]
        min_length = 20

        linear_sadf = get_sadf(log_prices, model='linear', add_const=True, min_length=min_length, lags=lags_int,
                               verbose=True)
        linear_sadf_no_const_lags_arr = get_sadf(log_prices, model='linear', add_const=False, min_length=min_length,
                                                 lags=lags_array, verbose=True)
        quadratic_sadf = get_sadf(log_prices, model='quadratic', add_const=True, min_length=min_length, lags=lags_int,
                                  verbose=True)

        sm_poly_1_sadf = get_sadf(log_prices, model='sm_poly_1', add_const=True, min_length=min_length, lags=lags_int,
                                  verbose=True)
        sm_poly_2_sadf = get_sadf(log_prices, model='sm_poly_2', add_const=True, min_length=min_length, lags=lags_int,
                                  verbose=True)
        sm_power_sadf = get_sadf(log_prices, model='sm_power', add_const=True, min_length=min_length, lags=lags_int,
                                 verbose=True)
        sm_exp_sadf = get_sadf(log_prices, model='sm_exp', add_const=True, min_length=min_length, lags=lags_int,
                               verbose=True)

        sm_power_sadf_phi = get_sadf(log_prices, model='sm_power', add_const=True, min_length=min_length, lags=lags_int,
                                     phi=0.5, verbose=True)
        sm_exp_sadf_phi = get_sadf(log_prices, model='sm_exp', add_const=True, min_length=min_length, lags=lags_int,
                                   phi=0.5, verbose=True)

        self.assertEqual(log_prices.shape[0] - min_length - lags_int - 1, sm_power_sadf.shape[0])  # -1 for series_diff
        self.assertEqual(log_prices.shape[0] - min_length - lags_int - 1, linear_sadf.shape[0])
        self.assertEqual(log_prices.shape[0] - min_length - lags_int - 1, quadratic_sadf.shape[0])
        self.assertEqual(log_prices.shape[0] - min_length - lags_int - 1, sm_poly_1_sadf.shape[0])
        self.assertEqual(log_prices.shape[0] - min_length - lags_int - 1, sm_poly_2_sadf.shape[0])
        self.assertEqual(log_prices.shape[0] - min_length - lags_int - 1, sm_exp_sadf.shape[0])
        self.assertEqual(log_prices.shape[0] - min_length - lags_int - 1, sm_exp_sadf_phi.shape[0])

        self.assertAlmostEqual(sm_power_sadf.mean(), 28.954, delta=1e-3)
        self.assertAlmostEqual(sm_power_sadf.iloc[29], 17.369, delta=1e-3)

        self.assertAlmostEqual(linear_sadf.mean(), -0.669, delta=1e-3)
        self.assertAlmostEqual(linear_sadf[29], -0.717, delta=1e-3)

        self.assertAlmostEqual(linear_sadf_no_const_lags_arr.mean(), 1.899, delta=1e-3)
        self.assertAlmostEqual(linear_sadf_no_const_lags_arr[29], 1.252, delta=1e-3)

        self.assertAlmostEqual(quadratic_sadf.mean(), -1.002, delta=1e-3)
        self.assertAlmostEqual(quadratic_sadf[29], -1.460, delta=1e-3)

        self.assertAlmostEqual(sm_poly_1_sadf.mean(), 26.033, delta=1e-3)
        self.assertAlmostEqual(sm_poly_1_sadf[29], 8.350, delta=1e-3)

        self.assertAlmostEqual(sm_poly_2_sadf.mean(), 26.031, delta=1e-3)
        self.assertAlmostEqual(sm_poly_2_sadf[29], 8.353, delta=1e-3)

        self.assertAlmostEqual(sm_exp_sadf.mean(), 28.916, delta=1e-3)
        self.assertAlmostEqual(sm_exp_sadf[29], 17.100, delta=1e-3)

        self.assertAlmostEqual(sm_power_sadf_phi.mean(), 1.4874, delta=1e-3)
        self.assertAlmostEqual(sm_power_sadf_phi.iloc[29], 2.4564, delta=1e-3)

        self.assertAlmostEqual(sm_exp_sadf_phi.mean(), 1.4787, delta=1e-3)
        self.assertAlmostEqual(sm_exp_sadf_phi[29], 2.4183, delta=1e-3)

        # Trivial series case.
        ones_series = pd.Series(index=log_prices.index, data=np.ones(shape=log_prices.shape[0]))
        trivial_sadf = get_sadf(ones_series, model='sm_power', add_const=True, min_length=min_length, lags=lags_int,
                                phi=0.5, verbose=True)

        self.assertTrue((trivial_sadf.unique() == [-np.inf]).all())  # All values should be -np.inf

        # Test rubbish model argument.
        self.assertRaises(ValueError, get_sadf, series=log_prices, model='rubbish_string', add_const=True,
                          min_length=min_length, lags=lags_int)

        # Assert that nans are parsed if singular matrix
        singular_matrix = np.array([[1, 0, 0], [-1, 3, 3], [1, 2, 2]])
        b_mean, b_var = get_betas(singular_matrix, singular_matrix)
        self.assertTrue(b_mean, [np.nan])
        self.assertTrue(b_var, [[np.nan, np.nan]])
